#ifndef CGAL_PARAMETRIC_CURVE_2_H
#define CGAL_PARAMETRIC_CURVE_2_H

#include "Types.h"
#include "Helpers.h"
#include "Curve_constants.h"
#include "Curve_2.h"
#include "Piecewise_lin_approximant_2.h"
#include "Default_FO.h"

//number of approximations points 
static const unsigned NUMBER = 70;
static const unsigned DEGREE = 1;

// Template class Parametric_curve_2 implements a generalized parametric
// curve in 2D f: [a, b] -> R^2.
// Template parameters:
//       NumType_      number type for coordinates
//       X_            pointer to function object which implements the x(t) 
//                     function
//       Y_            pointer to function object which implements the y(t) 
//                     function
// Inherits from the class Curve_2.

template <class NumType_, 
	  class X_=Default_FO<NumType_>* , 
	  class Y_=Default_FO<NumType_>* > 
class Parametric_curve_2 : 
    public Curve_2<NumType_>, 
    public geowin_update<std::list<CGAL::Point_2<CGAL::Cartesian<NumType_> > >,                         std::list<CGAL::Segment_2<CGAL::Cartesian<NumType_> > > > {
    //just for testing the private methods
    friend class ParamCurveTest;

public:    
    typedef X_ X;
    typedef Y_ Y;
    typedef std::list<NumType> ParamList;
    typedef ParamList::iterator ParamListIt;
    typedef Piecewise_lin_approximant_2<NumType> Lin_approximant; 
    typedef Lin_approximant::ParamVector ParamVector;   
    
         

    // init: X, Y, NumType, NumType -> void
    // Parameters: 
    //     x   pointer function object which implements the x(t) function 
    //         of the curve
    //     y   pointer function object which implements the y(t) function 
    //         of the curve
    //     a_  the low bound of t's range, t in [a_,b_] 
    //     b_  the up bound of t's range, t in [a_,b_]
    void init(const X x, const Y y, NumType a_, NumType b_) {
	assert(b_>a_);	        
	a = a_;
	b = b_;
	x_func = x;
	y_func = y;
    };

    // Parametric_curve_2: X, Y, NumType, NumType  -> Parametric_curve_2    
    // Parameters: 
    //     x   pointer function object which implements the x(t) function 
    //         of the curve
    //     y   pointer function object which implements the y(t) function 
    //         of the curve
    //     a_  the low bound of t's range, t in [a_,b_] 
    //     b_  the up bound of t's range, t in [a_,b_] 
    // Effect: contructs the Parametric_curve_2 object with x(t) = x and 
    //         y(t) = y whose parameter t is in the interval [a_,b_] default 
    //         value [LOW_BOUND, UP_BOUND] 
    //         and x(t) = x and y(t) = y    
    Parametric_curve_2(const X x, const Y y, NumType a_=LOW_BOUND, 
		       NumType b_=UP_BOUND) {
	init(x, y, a_, b_);	
	initial_approximation();	
    }

    // Copy constructor: Parametric_curve_2 -> Parametric_curve_2    
    // Parameters: 
    //      c   parametric curve
    // Effect: contructs a Parametric_curve_2 object which is a copy of c    
    Parametric_curve_2(const Parametric_curve_2&  c) {
	init(c.get_x_func(), c.get_y_func(), 
	     c.get_lower_bound(), c.get_upper_bound());
	lin_approx = c.get_lin_approximant();
	eps = c.get_tolerance();	
    }

    // Destructor
    virtual ~Parametric_curve_2() {
    }   
         
    // get_x_func: void -> X
    // Returns: pointer to the function x(t) of the curve     
    X get_x_func() const{
	return x_func;
    }

    // get_x_func: void -> Y
    // Returns: pointer to the function y(t) of the curve
    Y get_y_func() const{
	return y_func;
    }

    // get_lower_bound: void -> NumType
    // Returns: a where [a, b] is the domain of t
    NumType get_lower_bound() const {
	return a;
    } 

    // get_upper_bound: void -> NumType
    // Returns: b where [a, b] is the domain of t
    NumType get_upper_bound() const {
	return b;
    }
   
    // x: NumType -> NumType
    // Parameters:
    //      t   the parameter of the curve
    // Pre: a<=t<=b
    // Returns: the value x(t)   
    NumType x(NumType t) const {		
	assert (t>=a && t<=b);		
	return (*x_func)(t);
    }
    
    // y: NumType -> NumType
    // Parameters:
    //      t   the parameter of the curve
    // Pre: a<=t<=b
    // Returns: the value y(t) 
    NumType y(NumType t) const {	
	assert(t>=a && t<=b);
	return (*y_func)(t);	
    }    

    // calculate_point_at: NumType -> Point
    // Parameters:
    //      t   the parameter of the curve f: [a,b] -> R^2
    // Pre: a<=t<=b
    // Returns: a point on the curve with coordinates x(t), y(t)
    Point calculate_point_at(NumType t) const {
	Point p(x(t), y(t));	
	return p;
    }

    // get_lin_approximant: void -> Lin_approximant
    // Returns: the piecewise linear approximant of this curve
    Lin_approximant get_lin_approximant() const {	
	return lin_approx;
    }    
    
    // get_tolerance: void -> NumType
    // Returns: the tolerance of the current piecewise linear approximation
    //          eps. In the ideal case dist(p(t), lin_appr(t))<=eps
    //          for all t in [a,b]
    NumType get_tolerance() const {
	return eps;
    } 

    // get_name: void -> char*
    // Returns: the string "Parametric Curve"
    virtual char * get_name() const {
	return "Parametric Curve";
    }

    // update: PointList -> SegmentList
    // Parameters:
    //    in     container with the approximation points of this parametric 
    //           curve
    //    out    container with the segments linking the approximation points
    // Effect: called by every change of the input scene (the appr. points)
    //         and calculates the segments of the output scene (the curve)
    // Overwrites geowin_update<PointList, SegmentList>::update()
    virtual void update (const PointList& in, SegmentList& out) {	
	link_points_to_segments(in, out);	
    }

    // initial_approximation: unsigned -> void
    // Parameters:
    //     num   the number of segments in the linear approximant, 
    //           default value NUMBER    
    // Pre: num > 1
    // Effect: performes uniform parametrisation of the intervall [a, b]
    //         to create an initial piecewise linear approximant for 
    //         this curve    
    virtual void initial_approximation(unsigned num=NUMBER) const{
	Points appr_pts;
	ParamVector values;
	//a constant parametrisation step
	NumType step = (b-a)/((NumType) num);	
	NumType max_dist=0;
	NumType t, dist;
	for(unsigned i=0; i<num; i++) {	    
	    t = a+step*i;
	    appr_pts.push_back(calculate_point_at(t));	    
	    values.push_back(t);
	    if(i!=0) {
		Segment s(appr_pts[i-1], appr_pts[i]);		
		dist = squared_distance(calculate_point_at(
				    (values[i-1]+values[i])/(NumType) 2.0), s);
		if(dist>max_dist) max_dist=dist;
	    }
	}	
	appr_pts.push_back(calculate_point_at(b)); 
	values.push_back(b);	
	lin_approx.set_points(appr_pts.begin(), appr_pts.end(), 
			values.begin());	
	eps = _sqrt(max_dist);	
    }

    // refine_approximantion: NumType, bool -> void
    // Parameters:
    //    epsilon    the desired tolerance of the approximation
    //    use_deriv  if true, the approximation with the second derivative
    //                        is used
    //               else  approximation with testing in the middle is used  
    // Effect: refines the piecewise linear approximant of this curve using 
    //         adaptiv parametrisation step. In the ideal case 
    //         dist(lin_appr(t), p(t))<=epsilon for all t in [a,b]   
    virtual void refine_approximation(NumType epsilon=EPSILON, 
				      bool use_deriv=false) const{	
	if(use_deriv) second_deriv_approximation(epsilon);
	else insert_pt_in_middle_approximation(epsilon); 	
    }
    	
    // draw: void -> void 
    // Effect: opens a GeoWin window and draws the linear approximation of
    //         this parametric curve
    // Overwrites Curve_2::draw()
    virtual void draw() {
	if(eps > EPSILON) refine_approximation(EPSILON, false);

	geowin_init_default_type((PointList*)0, leda_string("ApprPointList"));
	geowin_init_default_type((SegmentList*)0, leda_string("SegmentList"));
	//creates a GW object for drawing the curve
	GeoWin GW("Parametric Curves Demo");

	//the approximation points are the input scene for drawing the curve
	PointList appr_pts(lin_approx.points_begin(),
			       lin_approx.points_end());
	geo_scene appr_scene = GW.new_scene(appr_pts);
	GW.set_point_style(appr_scene, leda_disc_point);
	GW.set_color(appr_scene, leda_blue);
	GW.set_visible(appr_scene, true);	

	draw(GW, appr_scene);
    }    
 
    // evaluate_derivative_at: NumType,int -> Vector, bool
    // Papameters:
    //    t      the parameter at which the derivative is computed
    //    r      compute the r-th derivative of the curve, default r=1
    //    deriv  will contain the r-th derivative of the curve at t, if 
    //           it exists
    // Returns: true, if the r-th derivative at t exists
    //          false otherwise
    // Pre:  a<=t<=b   
    // Effect: if the return value ist true, deriv contains the r-th 
    //         derivative of the curve at t.
    virtual bool evaluate_derivative_at(NumType t, Vector & deriv, 
					unsigned r=DEGREE) const {
	assert(a<=t && b>=t);
	assert(r>=1);	
	if(eps>0.01) {	    
	    refine_approximation(0.01);
	}	
	Lin_approximant deriv_approx;
	
	if(!lin_approx.discrete_derivative(deriv_approx, r)) 
	    return false;	
	deriv = (deriv_approx.evaluate_at(t) - CGAL::ORIGIN);
	return true;
    }

    // bbox: void -> Bbox    
    // Returns: the bounding box of this parametric curve (actualy of its
    //          linear approximant)
    virtual Bbox bbox() const {	
	return lin_approx.bbox(eps);
    };

    // has_on: Point, NumType -> bool
    // Parameters: 
    //     pt       the point is tested if it is on the curve
    //     epsilon  the upper bound for the distance from pt to the 
    //              approximation of the curve
    // Returns: true, if dist(pt, lin_approx)<=epsilon, which means that
    //                pt is at most 2*epsilon far way from this curve
    //          false otherwise
    // Tests if the point pt lies to within tolerance epsilon on the curve.
    bool has_on(const Point& pt, NumType epsilon=EPSILON) const {
	//lin_approx is empty
	if(lin_approx.get_number()==0) return false;	
	if(eps>epsilon) refine_approximation(epsilon);	
	if(!do_overlap(pt.bbox(), bbox())) return false;	
	return lin_approx.has_on(pt, epsilon);
    }
    
    // intersections: Parametric_curve_2 -> Points
    // Parameters: 
    //   curve         another parametric curve
    //   intersec_pts  the vector with the intersection points of the 
    //                 two curves 
    // Computes the intersection points of the piecewise linear approximants
    // of the curves
    template<class Xfunc, class Yfunc> 
    void intersections(Points & intersec_pts, 
		       const Parametric_curve_2<NumType, Xfunc, Yfunc> & curve,
		       NumType epsilon=EPSILON) const {
	intersec_pts.clear();
	if(lin_approx.get_number()==0 || 
	   (curve.get_lin_approximant()).get_number()==0) return;	
	if(eps>epsilon) refine_approximation(epsilon);	
	if(curve.get_tolerance()>epsilon) 
	   curve.refine_approximation(epsilon);	
	lin_approx.intersections(intersec_pts, curve.get_lin_approximant());
    }

protected:
    
    // Parametric_curve_2: void -> Parametric_curve_2    
    // Effect: contructs a Parametric_curve_2 object whose parameter is
    //         in the default interval [LOW_BOUND, UP_BOUND]
    // contructor without parameter    
    Parametric_curve_2(): a(LOW_BOUND), b(UP_BOUND), eps(1000.0), 
	                  lin_approx() {}   

    // draw: GeoWin, geo_scene, bool -> void 
    // Parameters:
    //    GW          a GeoWin object
    //    in_scene    the input scene with the approximation points
    //    do_edit     if true, draw calls GW.edit()
    // Effect: draws the piecewise linear approximant of this curve on GW
    virtual void draw(GeoWin& GW, geo_scene in_scene, 
		      bool do_edit = true) {	
			  //if(eps > EPSILON) refine_approximation();
	//lin_approx.draw();

	geo_scene result= GW.new_scene( *this, in_scene,
					 leda_string(get_name()));	
	GW.set_color(result, leda_blue);
	GW.set_line_width(result, 3);
	GW.set_visible(result, true);

	GW.display(80, 10);
	Bbox b = bbox();
	GW.init(b.xmin(), b.xmax(), b.ymin(), b.ymax(), 25);	
        //GW.edit(pts_scene);
	if(GW.get_grid_dist() > 1.0) GW.set_grid_dist(1.0);
	if(do_edit) GW.edit(result);
    };        

    // the domain of the parameter t, a<=t<=b
    NumType a, b;

    // the tolerance value used for the current linear approximantion
    // of this curve. In the ideal case is dist(lin_appr(t), p(t))<=eps
    // for all t in [a,b].
    mutable NumType eps;

    //the piecewise linear approximant of this parametric curve
    mutable Lin_approximant lin_approx;

private:
    //pointer to the function objects x and y,  t -> (x(t), y(t))
    X x_func;
    Y y_func;    

    // appr_segment_within_tolerance: NumType -> boolean
    // Parameter:
    //    t_source   the parameter value at the beginning of 
    //               the approximation segment 
    //    t_target   the parameter value at the end of the 
    //               approximation segment
    //    epsilon    the tolerance, dafault value EPSILON
    // Returns:
    //    true    if the distance from the curve at m=(t_source+t_target)/2
    //            to the middle of the segment of its linear approximant 
    //            is <= epsilon, dist(p(m), lin_appr(m))<=epsilon
    //    false   otherwise    
    // Determines whether a parametric curve hord can be approximated to 
    // within tolerance epsilon by the line segment joining the points at
    // t_source and t_target
    bool appr_segment_within_tolerance(NumType t_source, NumType t_target,
				       NumType epsilon=EPSILON) const {
	NumType e_2 = power(epsilon, 2);
	//the distance between P(t) and L(t) for t=(t_target + t_source)/2
	if(squared_distance(calculate_point_at((t_target+t_source)/(NumType) 2.0), 
			    midpoint(calculate_point_at(t_source), 
				     calculate_point_at(t_target)))> e_2) 
	     return false;
	else return true;
    }

    // insert_pt_in_middle_approximation: NumType -> void
    // Parameters:
    //    epsilon    the tolerance for this approximation. 
    // Effect: test the distance between the curve and its linear approximant
    //         in the middle m=(t_i+t_i+1)/2 of which interval [t_i, t_i+1] 
    //         and inserts points p(m) in the lin_approx while 
    //         dist(lin_appr(m), curve(m))>epsilon    
    void insert_pt_in_middle_approximation(NumType epsilon=EPSILON) const{
	ParamList values(lin_approx.parametrisation_begin(),
			   lin_approx.parametrisation_end());	
	if(values.size()<=1) return;
	NumType t;
	unsigned count = 0;
	unsigned depth = 0;

	ParamListIt prev = values.begin();
	ParamListIt curr = prev;
	curr++;
	// iterate through the parameter values of the lin. approximant
	// depth search
	while (curr != values.end() && depth<10) {		     
	    NumType left = *prev;
	    NumType right = *curr;
	    // if the approximation is not within tolerance epsilon
	    // take a new parameter value t=left+right/2 in the lin. 
	    // approximant and go on with the same parameter value (left)    
	    if(!appr_segment_within_tolerance(left, right, epsilon)) { 
		t = (left + right)/(NumType) 2.0;		
		values.insert(curr, t);
		curr = prev;
		count++;
		depth++;
	    } 
	    // else go on with the next parameter value
	    else {
		prev++;
		depth = 0;
	    }
	    curr++;
	}	
	// if any new parameter values was inserted update the 
	// linear approximant of this curve
	if (count > 0) {
	    Points appr_pts;
	    prev = values.begin();
	    while(prev!=values.end()) {
		appr_pts.push_back(calculate_point_at(*prev++));
	    }
	    ParamVector v(values.begin(), values.end());
	    lin_approx.set_points(appr_pts.begin(), appr_pts.end(), 
			v.begin());
	}
	eps = epsilon;		
    }

    // second_deriv_approximation: NumType -> void
    // Parameters:
    //    epsilon   the tolerance for this approximation
    // Effect: tests for which interval [t_i, t_i+1] if 
    //         p"(t)*(t_i+1 - t_i-1)/8 > epsilon is true and inserts points 
    //         p((t_i+t_i+1)/2) in the lin_approx 
    void second_deriv_approximation(NumType epsilon=EPSILON) const{
	PointList appr_pts(lin_approx.points_begin(),
			   lin_approx.points_end());
	ParamList values(lin_approx.parametrisation_begin(),
			   lin_approx.parametrisation_end());		
	if(values.size()<=1) return;	
	
	NumType e_2 = power(epsilon, 2);
	NumType t, coef_2, max_2, tmp_2, deriv_l_2, deriv_r_2;
	Lin_approximant second_deriv;		
	
	unsigned count = 0;	

	ParamListIt prev = values.begin();
	ParamListIt curr = prev;
	curr++;
	PointListIt prev_pt = appr_pts.begin();
	PointListIt curr_pt = prev_pt;
        curr_pt++;	
	
	bool ok = false;
	// wide search
	while(!ok) {
	    ok = true;	    
	    lin_approx.discrete_derivative(second_deriv, 2);
	    curr = prev = values.begin();	    
	    curr++;
	    curr_pt = appr_pts.begin();
	    curr_pt++;
	    if(prev != values.end())
		deriv_r_2 =(second_deriv.evaluate_at(*prev) - CGAL::ORIGIN) *
		    (second_deriv.evaluate_at(*prev) - CGAL::ORIGIN);
	    while ((curr != values.end())) {		     
		NumType left = *prev;
		NumType right = *curr;
		deriv_l_2 = deriv_r_2;
		deriv_r_2 =(second_deriv.evaluate_at(right) - CGAL::ORIGIN) *
		    (second_deriv.evaluate_at(right) - CGAL::ORIGIN);
		coef_2 = power((power((right-left), 2)/(NumType) 8.0),2);
		if(deriv_r_2 >= deriv_l_2)
		    tmp_2 = deriv_r_2*coef_2;
		else tmp_2 = deriv_l_2*coef_2;
		if(tmp_2 > e_2) {	 
		    ok = false;		    	    
		    t = (left + right)/(NumType) 2.0;		    
		    values.insert(curr, t);		    
		    appr_pts.insert(curr_pt, calculate_point_at(t));
		    if(tmp_2 > max_2) max_2 = tmp_2;	    
		}	       
		prev=curr++;
		curr_pt++;
	    } //end while curr!=end
	    if(!ok) {
		count++;
		ParamVector v(values.begin(), values.end());
		Points v_pt(appr_pts.begin(), appr_pts.end());
		lin_approx.set_points(v_pt.begin(), v_pt.end(),
				      v.begin());
	    }		    
	} //end while(!ok)	
	eps = epsilon;		
    }
};
#endif //CGAL_PARAMETRIC_CURVE_2_H
